"""Routes to get chat history"""
import json
import traceback

from fastapi import APIRouter, HTTPException, Response, Depends, Request, status
from promptengineers.fastapi.controllers import PromptController
from promptengineers.models.request import ReqBodyPromptSystem
from promptengineers.models.response import ResponsePromptSystemList, ResponsePromptSystem
from promptengineers.mongo.utils import JSONEncoder
from promptengineers.core.utils import logger
from promptengineers.core.exceptions import NotFoundException

router = APIRouter()
TAG = "Prompt"

def get_controller(request: Request) -> PromptController:
	try:
		return PromptController(request=request, user_repo=request.state.user_repo)
	except NotFoundException as e:
		# Handle specific NotFoundException with a custom message or logging
		logger.warn(f"Failed to initialize HistoryController: {str(e)}")
		raise HTTPException(status_code=404, detail=f"Initialization failed: {str(e)}") from e
	except Exception as e:
		# Catch all other exceptions
		logger.error(f"Unexpected error initializing HistoryController: {str(e)}")
		raise HTTPException(status_code=500, detail="Internal server error") from e

#################################################
# List Chat Histories
#################################################
@router.get(
	"/prompt/system",
	tags=[TAG],
	name='prompt_system_list',
	response_model=ResponsePromptSystemList
)
async def index(
	page: int = 1,
	limit: int = 50,
	controller: PromptController = Depends(get_controller),
):
	"""List resources"""
	try:
		result = await controller.index(page, limit)
		# Format Response
		data = json.dumps({
			'prompts': result,
		}, cls=JSONEncoder)
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error(err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[list_chat_histories]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Create Chat History
#################################################
@router.post(
	"/prompt/system",
	tags=[TAG],
	name='prompt_system_create',
	response_model=ResponsePromptSystem
)
async def create(
	body: ReqBodyPromptSystem,
	controller: PromptController = Depends(get_controller)
):
	"""Creates resource"""
	try:
		result = await controller.create(body)
		# Format Response
		data = json.dumps({
			**result
		})
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error(err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[create_chat_history]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Show Chat History
#################################################
@router.get(
	"/prompt/system/{prompt_id}",
	tags=[TAG],
	name='prompt_system_show',
	response_model=ResponsePromptSystem,
)
async def show(
	prompt_id: str,
	controller: PromptController = Depends(get_controller),
):
	"""Retrieve resource"""
	try:
		result = await controller.show(prompt_id)

		# Format Response
		data = json.dumps({
			**result
		}, cls=JSONEncoder)
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error("%s", err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[show_chat_history]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Update Chat History
#################################################
@router.put(
	"/prompt/system/{prompt_id}",
	tags=[TAG],
	name='prompt_system_update',
	response_model=ResponsePromptSystem,
)
async def update(
	prompt_id: str,
	body: ReqBodyPromptSystem,
	controller: PromptController = Depends(get_controller),
):
	"""Update resource"""
	try:
		await controller.update(prompt_id, body)
		data = json.dumps({
			'message': 'Chat history updated successfully.'
		})
		# Format Response
		return Response(status_code=200, content=data)
	except HTTPException as err:
		logger.error("%s", err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[update_chat_history]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Delete Chat History
#################################################
@router.delete(
	"/prompt/system/{prompt_id}",
	tags=[TAG],
	name='prompt_system_delete',
	status_code=status.HTTP_204_NO_CONTENT,
)
async def delete(
	prompt_id: str,
	controller: PromptController = Depends(get_controller),
):
	"""Delete Resource"""
	try:
		await controller.delete(prompt_id)
		# Format Response
		return Response(status_code=204)
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[delete_chat_history]: %s\n%s", err, tb)
		raise HTTPException(status_code=404, detail=str(err)) from err


"""Storage Service using MinIO"""
import os
from minio import Minio
from minio.error import S3Error

from promptengineers.core.config import S3_REGION, MINIO_SERVER
from promptengineers.core.utils import logger


class StorageService:
	"""Storage Service Class for MinIO"""
	def __init__(self, access_key_id, secret_access_key, minio_server: str = None):
		self.minio_server = minio_server if minio_server else (MINIO_SERVER or f"s3.{S3_REGION}.amazonaws.com")
		self.access_key_id = access_key_id
		self.secret_access_key = secret_access_key
		self.client = Minio(
			endpoint=self.minio_server,
			access_key=self.access_key_id,
			secret_key=self.secret_access_key,
			secure=False if MINIO_SERVER or minio_server else True
		)

	def retrieve_all_files_raw(self, bucket: str, prefix: str = ''):
		"""Retrieve all file details from a bucket"""
		files = []
		try:
			objects = self.client.list_objects(bucket, prefix=prefix, recursive=True)
			for obj in objects:
				files.append(obj)
		except S3Error as err:
			logger.error("Error retrieving files: %s", err)
		return files

	def retrieve_all_files(self, bucket: str, prefix: str = ''):
		"""Retrieve all filenames from a bucket"""
		files = []
		try:
			objects = self.client.list_objects(bucket, prefix=prefix, recursive=True)
			for obj in objects:
				filename = obj.object_name.split('/')[-1]
				if filename:
					files.append(filename)
		except S3Error as err:
			logger.error("Error retrieving filenames: %s", err)
		return files

	def retrieve_file(self, bucket: str, path: str):
		"""Retrieve a file from a bucket"""
		try:
			response = self.client.get_object(bucket, path)
			return response.data
		except S3Error as err:
			logger.error("Error retrieving file %s: %s", path, err)
			return None

	def delete_file(self, bucket: str, path: str):
		"""Delete a file from a bucket"""
		try:
			self.client.remove_object(bucket, path)
		except S3Error as err:
			logger.error("Error deleting file %s: %s", path, err)
			raise ValueError(f"Failed to delete file: {err}") from err

	def upload_file(self, file_name, bucket, object_name=None):
		"""Upload a file to a MinIO bucket

		:param file_name: File to upload
		:param bucket: Bucket to upload to
		:param object_name: MinIO object name. If not specified then file_name is used
		:return: True if file was uploaded, else False
		"""

		if object_name is None:
			object_name = os.path.basename(file_name)

		try:
			with open(file_name, 'rb') as file_data:
				file_stat = os.stat(file_name)
				self.client.put_object(
					bucket, object_name, file_data, file_stat.st_size
				)
		except S3Error as err:
			logger.error("Error uploading file %s: %s", file_name, err)
			return False
		return True



"""Functions configuration file."""

from promptengineers.functions import get_word_length

AVAILABLE_FUNCTIONS = {
	"get_word_length": get_word_length
}

FUNCTIONS = [
	{
		"name": "get_word_length",
		"description": "Returns the length of a word.",
		"parameters": {
			"type": "object",
			"properties": {
				"word": {
					"type": "string",
					"description": "The word whose length needs to be determined."
				}
			},
			"required": ["word"]
		}
	}
]



"""Utilites for Chains"""
from langchain.agents import load_tools
from langchain.tools import AIPluginTool
from langchain_core.documents.base import Document
from langchain.agents.agent_toolkits import create_retriever_tool

from promptengineers.retrieval.strategies import VectorstoreContext
from promptengineers.core.config.tools import AVAILABLE_TOOLS
from promptengineers.tools.utils import filter_tools

def get_chat_history(inputs: tuple) -> str:
    """Formats the chat history into a readable format for the chatbot"""
    res = []
    for human, assistant in inputs:
        res.append(f"Human: {human}\nAssistant: {assistant}")
    return "\n".join(res)

def combine_documents(documents: list[Document]):
    combined_content = ""
    for doc in documents:
        combined_content += doc.page_content + "\n"
    return combined_content

def retrieve_system_message(messages):
    """Retrieve the system message"""
    try:
        filtered_messages = list(filter(lambda message: message['role'] == 'system', messages))
        return filtered_messages[0]['content']
    except IndexError:
        return None

def retrieve_chat_messages(messages):
    """Retrieve the chat messages"""
    return [
        (msg["content"]) for msg in messages if msg["role"] in ["user", "assistant"]
    ]

def gather_tools(
    tools: list[str] = None,
    available_tools: dict[str, any] = None,
    vectorstore: VectorstoreContext = None,
    plugins: list[str] = None,
):
    """Gather tools from the tools list"""
    filtered_tools = filter_tools(tools or [], available_tools or AVAILABLE_TOOLS)

    ## Add docs tool
    if vectorstore:
        docs_tool = create_retriever_tool(
            vectorstore.as_retriever(),
            "search_docs",
            """
            Searches and returns relevant documents. It is a requirement to use this for every query.
            Rewrite the user to be a detailed question.
            """,
        )
        filtered_tools.append(docs_tool)

    ## Add plugins
    if plugins and len(plugins) > 0:
        loaded_tools = load_tools(["requests_all"])
        for tool in plugins:
            tool = AIPluginTool.from_plugin_url(tool)
            loaded_tools += [tool]
        filtered_tools += loaded_tools

    return filtered_tools


"""Chain Service"""
from langchain.agents import AgentExecutor, OpenAIFunctionsAgent
from langchain.agents.openai_functions_agent.agent_token_buffer_memory import AgentTokenBufferMemory
from langchain.callbacks.base import BaseCallbackHandler
from langchain.chains import (
    ConversationChain,
	ConversationalRetrievalChain,
	LLMChain
)
from langchain.memory import ConversationBufferMemory, ConversationSummaryBufferMemory
from langchain.prompts import MessagesPlaceholder
from langchain.schema import SystemMessage
from langchain.chains.chat_vector_db.prompts import QA_PROMPT
from langchain.chains.question_answering import load_qa_chain
from langchain.prompts import (
    ChatPromptTemplate,
    MessagesPlaceholder,
    HumanMessagePromptTemplate,
)
from promptengineers.core.config import APP_ENV
from promptengineers.prompts.templates import get_system_template, get_retrieval_template


class ChainService:
	"""Chain Service"""
	def __init__(self, model):
		self.model = model

	#########################################################
	## Conversation
	#########################################################
	def conversation(self):
		prompt_template = ChatPromptTemplate.from_messages(
			[
				MessagesPlaceholder(variable_name="context"),
				HumanMessagePromptTemplate.from_template("{input}")
			]
		)
		memory = ConversationBufferMemory(return_messages=True, memory_key="context")
		llm_chain = ConversationChain(llm=self.model, prompt=prompt_template, memory=memory, verbose=False)

		return llm_chain

	#########################################################
	## Question Answering
	#########################################################
	def condense_question(self, system_message):
		"""Condense a question into a single sentence."""
		return LLMChain(
			llm=self.model,
			prompt=get_system_template(system_message),
		)

	def collect_docs(self, system_message):
		"""Collect documents from the vectorstore."""
		return load_qa_chain(
			self.model,
			chain_type='stuff',
			prompt=get_system_template(system_message)
		)
	
	def conversation_retrieval(
		self,
		vectorstore,
		system_message,
		chat_history,
		callbacks = None
	):
		"""Retrieve a conversation."""
		memory = ConversationSummaryBufferMemory(llm=self.model,
												memory_key="chat_history",
												return_messages=True,
												output_key='answer')
		for message in chat_history:
			if message[0] and message[1]:
				memory.chat_memory.add_user_message(message[0])
				memory.chat_memory.add_ai_message(message[1])
			else:
				memory.chat_memory.add_user_message(message[0])
		return ConversationalRetrievalChain.from_llm(
			llm=self.model,
			condense_question_llm=self.model,
			retriever=vectorstore.as_retriever(),
			memory=memory,
			combine_docs_chain_kwargs={"prompt": get_retrieval_template(system_message)},
			return_source_documents=True,
			verbose=True if APP_ENV == 'local' or APP_ENV == 'development' else False,
			callbacks=callbacks or []
		)


	#########################################################
	## Agent
	#########################################################
	def create_executor(
		self,
		content,
		tools,
		chat_history,
		verbose=True if APP_ENV == 'local' or APP_ENV == 'development' else False,
		return_messages = True,
		callbacks = [],
		return_intermediate_steps = True
	):
		system_message = SystemMessage(content=content)
		prompt = OpenAIFunctionsAgent.create_prompt(
			system_message=system_message,
			extra_prompt_messages=[MessagesPlaceholder(variable_name="chat_history")]
		)
		memory = AgentTokenBufferMemory(memory_key="chat_history", 
								  		llm=self.model, 
										return_messages=return_messages)
		if len(chat_history) > 0:
			for message in chat_history:
				if message[0] and message[1]:
					memory.chat_memory.add_user_message(message[0])
					memory.chat_memory.add_ai_message(message[1])
				else:
					memory.chat_memory.add_user_message(message[0])
		agent = OpenAIFunctionsAgent(llm=self.model, tools=tools, prompt=prompt)
		return AgentExecutor(
			agent=agent,
			tools=tools,
			memory=memory,
			verbose=verbose,
			callbacks=callbacks,
			return_intermediate_steps=return_intermediate_steps
		)

	def agent_with_tools(
			self,
			system_message: str,
			chat_history,
			tools,
			callbacks: list[BaseCallbackHandler] = None
		):
		"""Agent with tools."""
		agent_executor = self.create_executor(system_message, tools, chat_history, callbacks=callbacks)
		return agent_executor


"""Pinecone Service"""
from typing import List

import pinecone
from langchain.vectorstores import Pinecone

from promptengineers.retrieval.utils import split_docs

class PineconeService:
    def __init__(self, api_key: str, env: str, index_name: str):
        self.api_key = api_key
        self.env = env
        self.index_name = index_name


    def client(self):
        pinecone.init(api_key=self.api_key, environment=self.env or 'us-east1-gcp')
        return pinecone


    #############################################################
    ## List Indexes
    #############################################################
    def list_indexes(self):
        return self.client().list_indexes()


    #############################################################
    ## Create or Get Index
    #############################################################
    def index(
		self,
		metric: str = 'cosine',
		dimension: int = 1536
	):
        if self.index_name not in self.client().list_indexes():
			# we create a new index
            self.client().create_index(
				name=self.index_name,
				metric=metric,
				dimension=dimension
			)
        return self.client().Index(index_name=self.index_name)



    #############################################################
    ## Describe Index Stats
    #############################################################
    def describe_index_stats(self):
        return self.index().describe_index_stats()


    #############################################################
    ## Delete Vectors
    #############################################################
    def delete(
		self,
		namespace: str,
		ids: List[str] = None,
	):
        if self.index_name in self.client().list_indexes():
            delete_all = True if not ids else False
            deleted = self.index().delete(
                ids=ids,
                delete_all=delete_all,
                namespace=namespace
            )
            return deleted
        else:
            return False


    #############################################################
    ## Create Vectorstore from Documents
    #############################################################
    def from_existing(
		self,
		embeddings,
		namespace: str,
	):
        self.client()
        return Pinecone.from_existing_index(
            index_name=self.index_name,
            embedding=embeddings,
            namespace=namespace
        )


    #############################################################
    ## Create Vectorstore from Documents
    #############################################################
    def from_documents(
		self,
		loaders,
		embeddings,
		namespace: str,
		chunk_size: int = 1000,
		chunk_overlap: int = 100,
	):
        docs = []
        for loader in loaders:
            docs.extend(loader.load())
        self.client()
        return Pinecone.from_documents(
            split_docs(docs, chunk_size, chunk_overlap),
            embeddings,
            index_name=self.index_name,
            namespace=namespace
        )





from promptengineers.core.interfaces.repos import IUserRepo
from promptengineers.core.validations import Validator
from promptengineers.repos.user import UserRepo
from promptengineers.retrieval.strategies import PineconeStrategy, RedisStrategy

class RetreivalFactory:
	def __init__(
			self,
			provider: ('redis', 'pinecone'), 
			index_name: str,
			embeddings,
			user_id: str,
			user_repo: IUserRepo = UserRepo()
	):
		self.provider = provider
		self.index_name = index_name
		self.embeddings = embeddings
		self.user_id = user_id
		self.user_repo = user_repo
		self.validator = Validator()

	def __call__(self) -> PineconeStrategy | RedisStrategy:
		if self.provider in 'pinecone':
			required_keys = ['PINECONE_API_KEY', 'PINECONE_ENV', 'PINECONE_INDEX']
			tokens = self.user_repo.find_token(self.user_id, required_keys)
			self.validator.validate_api_keys(tokens, required_keys)
			vectorstore_strategy = PineconeStrategy(
				embeddings=self.embeddings,
				api_key=tokens.get(required_keys[0]),
				env=tokens.get(required_keys[1]),
				index_name=tokens.get(required_keys[2]),
				namespace=self.index_name,
			)
		elif self.provider in 'redis':
			required_keys = ['REDIS_URL']
			tokens = self.user_repo.find_token(self.user_id, required_keys)
			self.validator.validate_api_keys(tokens, required_keys)
			vectorstore_strategy = RedisStrategy(
				redis_url=tokens.get(required_keys[0]),
				index_name=self.index_name,
				embeddings=self.embeddings,
			)
		else:
			raise ValueError(f"Invalid index provider {self.provider}")
		return vectorstore_strategy


from typing import List
from pydantic import BaseModel

from promptengineers.models.settings import ChatSettingDocument, CHAT_SETTING_DOCUMENT

class ResponseSettingsList(ChatSettingDocument): # pylint: disable=too-few-public-methods
	settings: List[ChatSettingDocument] = []

	class Config: # pylint: disable=too-few-public-methods
		"""Return list of agent settings"""
		json_schema_extra = {
			"example": {
				"settings": [CHAT_SETTING_DOCUMENT]
			}
		}

class ResponseSetting(ChatSettingDocument): # pylint: disable=too-few-public-methods
	setting: ChatSettingDocument

	class Config: # pylint: disable=too-few-public-methods
		"""Return single agent settings"""
		json_schema_extra = {
			"example": {
				"setting": CHAT_SETTING_DOCUMENT
			}
		}




from promptengineers.models.settings import ChatSettings, CHAT_SETTINGS

class ReqBodySettings(ChatSettings): # pylint: disable=too-few-public-methods

	class Config: # pylint: disable=too-few-public-methods
		"""Request body for settings"""
		json_schema_extra = {
			"example": CHAT_SETTINGS
		}


from bson.objectid import ObjectId

from promptengineers.core.interfaces.controllers import IController
from promptengineers.core.interfaces.repos import IUserRepo
from promptengineers.repos.user import UserRepo
from promptengineers.mongo.service import MongoService

class HistoryController(IController):
	def __init__(
		self,
		user_id: str = None,
		request = None,
		user_repo: IUserRepo = None,
		db_name: str = None,
		col_name: str = None
	):
		self.request = request
		self.user_id = user_id or getattr(request.state, "user_id", None)
		self.user_repo = user_repo or UserRepo()
		self.history_service = MongoService(
			host=self.user_repo.find_token(self.user_id, 'MONGO_CONNECTION'),
			db=db_name or self.user_repo.find_token(self.user_id, 'MONGO_DB_NAME'),
			collection=col_name or 'history'
		)

	##############################################################
	### Create Chat History
	##############################################################
	async def index(self, page: int = 1, limit: int = 10, **kwargs):
		# Base query with user_id
		query = {'user_id': ObjectId(self.user_id)}

		# Add additional filters passed as keyword arguments
		for key, value in kwargs.items():
			if value is not None:  # Ensure that a value is provided for the filter
				# Convert to ObjectId if necessary, for example, if key ends with "_id"
				if key in {'setting'}:
					query[key] = ObjectId(value)
				else:
					query[key] = value

		# Get the result from the history_service
		result = await self.history_service.list_docs(query, limit, page)
		return result

	##############################################################
	### Create Chat History
	##############################################################
	async def create(
			self, 
			body, 
			keys: set[str] = {'setting', 'messages', 'tags', 'title'}
		):
		body = await self.request.json() if self.request else body
		body = dict((k, body[k]) for k in keys if k in body)
		body['user_id'] = ObjectId(self.user_id)
		if body.get('setting', False):
			body['setting'] = ObjectId(body['setting'])
		result = await self.history_service.create(dict(body))
		return result

	##############################################################
	### Show Chat History
	##############################################################
	async def show(self, id: str = None, query = None):
		result = await self.history_service.read_one(
			query or {'_id': ObjectId(id), 'user_id': ObjectId(self.user_id)}
		)
		return result

	##############################################################
	### Update Chat History
	##############################################################
	async def update(
			self, 
			id: str, 
			body: any, 
			keys: set[str] = {'setting', 'messages', 'tags', 'title'}
		):
		body = await self.request.json() if self.request else body
		body = dict((k, body[k]) for k in keys if k in body)
		if body.get('setting', False):
			body['setting'] = ObjectId(body['setting'])
		result = await self.history_service.update_one(
			{'_id': ObjectId(id), 'user_id': ObjectId(self.user_id)},
			dict(body)
		)
		return result

	##############################################################
	### Delete Chat History
	##############################################################
	async def delete(self, id: str):
		result = await self.history_service.delete_one(
			{'_id': ObjectId(id), 'user_id': ObjectId(self.user_id)}
		)
		return result



"""Configuration files for the project."""
import os

# Path to the vector store
APP_ENV = os.getenv("APP_ENV", 'development')
APP_NAME = os.getenv("APP_NAME", 'Prompt Engineers AI - API Server')
APP_SECRET = os.getenv("APP_SECRET", '')
APP_VERSION = os.getenv("APP_VERSION", '')
APP_ORIGINS = os.getenv("APP_ORIGINS", '*')

# OpenAI API Key
OPENAI_API_KEY = os.getenv("OPENAI_API_KEY", '')
## Ollama URL
OLLAMA_BASE_URL= os.getenv("OLLAMA_BASE_URL", 'http://localhost:11434')

# Mongo
DB_NAME= os.getenv('DB_NAME', 'precisionx-dev')
DB_COLLECTION = os.getenv('DB_COLLECTION', 'history')
MONGO_CONNECTION = os.getenv("MONGO_CONNECTION", f'mongodb://localhost:27017')
MONGO_DB_NAME = os.getenv("MONGO_DB_NAME", f'llm-server')

# Redis
REDIS_URL = os.getenv("REDIS_URL", 'redis://localhost:6379/0')

# S3 Bucket Credentials
BUCKET = os.getenv("BUCKET", 'precision-x')
S3_REGION = os.getenv("S3_REGION", 'us-east-1')
ACCESS_KEY_ID = os.getenv("ACCESS_KEY_ID", '')
ACCESS_SECRET_KEY = os.getenv("ACCESS_SECRET_KEY", '')
MINIO_SERVER = os.getenv("MINIO_SERVER", '')

# Pinecone Credentials
PINECONE_API_KEY = os.getenv("PINECONE_API_KEY", '')
PINECONE_ENV = os.getenv("PINECONE_ENV", '')
PINECONE_INDEX = os.getenv("PINECONE_INDEX", '')

# Blockchain Credentials
ALCHEMY_API_KEY = os.getenv("ALCHEMY_API_KEY", '')
